/*
 *
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Copyright (c) 2024-2025, Arm Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice, this
 *    list of conditions and the following disclaimer in the documentation and/or
 *    other materials provided with the distribution.
 * - Neither the name of ARM nor the names of its contributors may be used to
 *    endorse or promote products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "system_spinlock.h"
#include "acamera_logger.h"
#include "acamera_types.h"

#include "FreeRTOS.h"
#include "semphr.h"

#ifndef CMSIS_device_header
/* CMSIS pack default header, containing the CMSIS_device_header definition */
#include "RTE_Components.h"
#endif
#include CMSIS_device_header

#include "cmsis_compiler.h"

//! Test if in interrupt mode
__STATIC_INLINE
int isInterrupt()
{
    return (SCB->ICSR & SCB_ICSR_VECTACTIVE_Msk) != 0;
}

int system_spinlock_init(sys_spinlock *lock)
{
    SemaphoreHandle_t xSemaphore = xSemaphoreCreateMutex();
    if (xSemaphore == NULL) {
        return -1;
    }
    *lock = xSemaphore;
    return 0;
}

unsigned long system_spinlock_lock(sys_spinlock lock)
{
    LOG(LOG_DEBUG, "LOCK 0x%x", (uint32_t)lock);
    if (isInterrupt()) {
        if (xSemaphoreTakeFromISR((SemaphoreHandle_t)lock, NULL) != pdTRUE) {
            LOG(LOG_DEBUG, "    FAIL");
            return -1;
        }
    } else {
        if (xSemaphoreTake((SemaphoreHandle_t)lock, portMAX_DELAY) != pdTRUE) {
            LOG(LOG_DEBUG, "    FAIL");
            return -1;
        }
    }
    return 0;
}

void system_spinlock_unlock(sys_spinlock lock, unsigned long flags)
{
    LOG(LOG_DEBUG, "UNLOCK 0x%x", (uint32_t)lock);
    if (isInterrupt()) {
        if (xSemaphoreGiveFromISR((SemaphoreHandle_t)lock, NULL) != pdTRUE) {
            LOG(LOG_DEBUG, "    FAIL");
            return;
        }
    } else {
        if (xSemaphoreGive((SemaphoreHandle_t)lock) != pdTRUE) {
            LOG(LOG_DEBUG, "    FAIL");
            return;
        }
    }
}

void system_spinlock_destroy(sys_spinlock lock)
{
    vSemaphoreDelete((SemaphoreHandle_t)lock);
}
